/*--------------------------------------------------------------------------*/
/*
 * This program contains code originally written by Toshiki Fujisawa,
 * Copyright (C) 2000.
 * Copyright (C) 2001 by Kazuo Nakayoshi <kazuo.nakayoshi@kek.jp>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*--------------------------------------------------------------------------*/
/*
 * USB Device Driver for Logitec LUB-SC and Kinetics 3929
 *
 *   Logitec LUB-SC : SCSI-USB Adapter Cable
 *   Kinetics 3929  : CAMAC SCSI Crate Controller
 *
 *   3 May, 2001 First release(K)
 */
/*--------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/malloc.h>
#include <linux/usb.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
//#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/proc_fs.h>

#include "xcamac.h"
/*
#define USB_XCAMAC_MINOR_BASE   129   
//#define MAX_TRANS_BYTES       16384 
#define MAX_TRANS_BYTES    1024*256
#define VID_PID_LOGITEC   0x7890004     
#define VID_PID_ADAPTEC   0x3f32000     
#define SCSI_ID                   0
#define SENSE_KEY_WRITE_RETRY  0x06
#define SENSE_KEY_READ_RETRY   0x06
#define CAMACW  1
#define CAMACL  2
#define RMINH   3
#define BLOCKW  4
#define BLOCKL  5
#define STOP    0x00
#define IGNORE  0x08
#define REPEAT  0x10
#define SCAN    0x18

#define TEST    0x00
#define SENSE   0x03
#define RECV    0x08
#define SEND    0x0A
#define INQUIRY 0x12
#define DIAG    0x1D

#define LAM_PAT 0x06
#define SINGLE  0x09
#define SETUP   0x0C
#define REG_ACC 0x0D

#define DSNDCLR(x) 
#define DSNDCDB(x) 
#define DRCVDAT(x) 
#define DSNODAT(x) 
#define DSREAD(x)  
#define DCREAD(x)  
#define DSSENS(x) 
#define DSNDDAT(x) 
#define DSWRITE(x) 
#define DCHKCAM(x) 
#define DXIOCTL(x) 
#define DXOPEN(x)  
#define DPROBE(x)  x
#define DSSNDID(x)

//#define TIMEMEASURE
*/

static int scsi_id;
MODULE_PARM(scsi_id, "i");

/*--------------------------------------------------------------------------*/
struct xcamac_data{
	struct usb_device       *dev;
	int                     is_open;
	int                     is_busy;
	int                     present;
	wait_queue_head_t       wait_queue;
	int                     ep_inp;
	int                     ep_out;
        int                     scsi_id;
	int                     sense_key;
	int                     sense_code;
	int                     sense_qual;
};

static struct xcamac_data *global_xcamac;
/*
struct camac_single{
	unsigned int naf;
	unsigned int data;
	unsigned int qx;
};

struct camac_block{
	unsigned int naf;
	unsigned int mode;
	unsigned int ch;
	void *data;
	unsigned int retlen;
};
*/
/*--------------------------------------------------------------------------*/
typedef __u8 err_status_t;
/*--------------------------------------------------------------------------*/
/*
  NAF High Byte 
  7  6  5   4   3   2   1   0(bit)
  0  0  N16 N8  N4  N2  N1  A8

  NAF Low Byte
  7   6   5   4    3   2   1   0(bit)
  A4  A2  A1  F16  F8  F4  F2  F1
 */
/*--------------------------------------------------------------------------*/
#define DEBUG_XCAMAC
/*--------------------------------------------------------------------------*/
/*
static int
send_clear_feature(struct file *file)
{
	struct usb_device *udev;
	unsigned int       pipe;
	int                r;

	DSNDCLR(printk("****** in send_clear_feature\n"));

	udev   = global_xcamac->dev;
	pipe   = usb_sndctrlpipe(udev, 0);

	DSNDCLR(printk("****** usb_clear_halt\n"));
	r = usb_clear_halt(udev, pipe);


	if( r < 0 ){
		printk("****** clear_feature err=%d\n", r );
		return -1;
	}
	DSNDCLR(printk("****** out send_clear_feature\n"));
	return 0;               
}
*/
/*--------------------------------------------------------------------------*/
static int
send_cdb(struct file *file, unsigned char *cdb, unsigned short size)
{
	struct usb_device *udev;
	unsigned int       pipe;
	__u8               requesttype = 0x21;
	__u8               request     = 0x00;
	__u16              value       = 0x0000;
	__u16              index       = 0x0000;
	//__u16              size;
	//int                timeout     = 60*HZ; /* 5sec */
	int                timeout     = 10*HZ; /* 5sec */
	int                r;

#ifdef TIMEMEASURE
	struct timeval tv1, tv2;
	long sec, usec;
#endif

	DSNDCDB(printk("**** in send_cdb\n"));

	udev = global_xcamac->dev;
	pipe   = usb_sndctrlpipe(udev, 0);

#ifdef TIMEMEASURE
	do_gettimeofday(&tv1);
#endif

	DSNDCDB(printk("**** send_cdb usb_control_msg\n"));
	r = usb_control_msg(udev, pipe, request, requesttype,
		      value, index, cdb, size, timeout);
#ifdef TIMEMEASURE
	do_gettimeofday(&tv2);
#endif

	if( r < 0 ){
		printk("**** send_cdb r = %d\n",r);
		return -1;
	}

#ifdef TIMEMEASURE
	//sec = (tv2.tv_sec) - (tv1.tv_sec);
	usec = (tv2.tv_usec) - (tv1.tv_usec);

	//printk("**** send cdb sec  = %ld sec\n", sec);
	printk("**** send cdb usec = %ld usec\n", usec);
#endif

	DSNDCDB(printk("**** r = %d\n", r));
	DSNDCDB(printk("**** out send_cdb\n"));
	return r;
}
/*--------------------------------------------------------------------------*/
static int
recv_data(struct file *file, void *buf, int len)
{
	struct usb_device     *udev;
	unsigned int           pipe;
	int                    r;
	int                    ans_len;
	int                    try;
	const int              max_try = 10;
	struct xcamac_data     *xcamac = file->private_data;

	udev   = xcamac->dev;
	pipe   = usb_rcvbulkpipe(udev, xcamac->ep_inp);

	DRCVDAT(printk("******* in recv_data\n"));

	for( try = 0; try < max_try; try++) {
		DRCVDAT(printk("******* try = %x\n",try));
		if(signal_pending(current)){
			xcamac->is_busy = 0;
			printk("ERR pending\n");
			return -EINTR;
		}
		if( udev == NULL ){
			xcamac->is_busy = 0;
			printk("ERR NODEV\n");
			return -ENODEV;
		}
		if( (r=usb_bulk_msg(udev, pipe, buf,
			len, &ans_len, 10*HZ)) >= 0 ){
			DRCVDAT(printk("******* recv_dat: r = %d (0x%x)\n",r,r));
			DRCVDAT(printk("******* ans_len : r = %d (0x%x)\n", ans_len, ans_len));
			return ans_len;
		}
		if( r != USB_ST_TIMEOUT ){
			DRCVDAT(printk("******* ERR!! not USB_ST_TIMEOUT\n"));
			//return -EIO;
			continue;
		}
		DRCVDAT(printk("******* OK USB_ST_TIMEOUT wait ...\n"));

		//interruptible_sleep_on (&xcamac->wait_queue);
		//interruptible_sleep_on_timeout (&xcamac->wait_queue, 1*HZ );
		//interruptible_sleep_on_timeout (&xcamac->wait_queue, 2*HZ );
	}
	DRCVDAT(printk("******* ERR read retry over\n"));

	xcamac->is_busy = 0;

	DRCVDAT(printk("******* out recv_data\n"));
	return -EIO;
}
/*--------------------------------------------------------------------------*/
static unsigned int
scsi_no_dat(struct file *file, unsigned char *cdb, unsigned short cdb_size)
{
	int     r;

	DSNODAT(printk("*** in scsi_no_dat\n"));

	if( (r = send_cdb(file, cdb, cdb_size)) < 0 )
		return -EFAULT;
	
	DSNODAT(printk("*** out scsi_no_dat\n"));
	return 0;
}
/*--------------------------------------------------------------------------*/
static unsigned int
scsi_read(struct file *file, unsigned char *cdb, unsigned short cdb_size, void *buf, int len)
{
	int     r;
	int     rest_len, cur_len;

#ifdef TIMEMEASURE
	struct timeval tv1, tv2;
	long sec, usec;
#endif

	DSREAD(printk("*** in scsi_read\n"));
	DSREAD( for(i = 0; i<cdb_size; i++) printk("send_cdb[%d] = %x\n", i, *(cdb + i)););


	if( (r = send_cdb(file, cdb, cdb_size)) < 0 ){
		return -EFAULT;
	}


	rest_len = len;

#ifdef TIMEMEASURE
	do_gettimeofday(&tv1);
#endif

	while( rest_len > 0 ){
		/*
		  cur_len = ( MAX_TRANS_BYTES < rest_len )
		  ? MAX_TRANS_BYTES : rest_len ;
		*/
		if(MAX_TRANS_BYTES < rest_len){
			cur_len = MAX_TRANS_BYTES;
			printk("MAX_TRANS_BYTES\n");
		}else 
			cur_len = rest_len;

		DSREAD(printk("*** rest_len = %x\n", rest_len));
		if( (r=recv_data(file, buf, cur_len )) < 0 ){
			printk("*** (scsi_read)ERR r = %x\n", r);
			//return r;
		}
		DSREAD(printk("*** (scsi_read)r = %x\n", r));
		
		buf      += cur_len;
		rest_len -= cur_len;
	}

#ifdef TIMEMEASURE
	do_gettimeofday(&tv2);
	//sec = (tv2.tv_sec) - (tv1.tv_sec);
	usec = (tv2.tv_usec) - (tv1.tv_usec);
	printk("*** recv dat usec = %ld usec\n", usec);
#endif

	DSREAD(printk("*** out scsi_read\n"));
	return len;
}
/*--------------------------------------------------------------------------*/
static unsigned int
scsi_request_sense(struct file *file)
{
	size_t len = 0x12;
	unsigned char cdb_size = 0x06;
	unsigned char cdb_request_sense[0x06] = {
		/* 0x00 */ 0x03,        /* SCSI REQUEST SENSE COMMAND */
		/* 0x01 */ 0x00,
		/* 0x02 */ 0x00,        
		/* 0x03 */ 0x00,        
		/* 0x04 */ 0x12,        
		/* 0x05 */ 0x00
	};

	unsigned char buf[0x12];
	struct xcamac_data *xcamac = file->private_data;

	DSSENS(printk("***** in scsi_request_sense\n"));
	if( scsi_read(file, cdb_request_sense, cdb_size, (unsigned int *)buf, len ) != len ){
		printk("request_sense fail\n");
		return -1;
	}

	DSSENS(printk("***** scsi_request_sense sense_key  = %x\n", buf[ 2]&0x0f));
	DSSENS(printk("***** scsi_request_sense sense_code = %x\n", buf[12]&0xff));
	DSSENS(printk("***** scsi_request_sense sense_qual = %x\n", buf[13]&0xff));
  
	xcamac->sense_code = buf[12] & 0xff;
	xcamac->sense_key  = buf[2]  & 0x0f;
	xcamac->sense_qual = buf[13] & 0xff;

	return buf[0x02];
}
/*--------------------------------------------------------------------------*/
static int
send_data(struct file *file, void *buf, int len)
{
	struct usb_device *udev;
	unsigned int       pipe;
	int                r;
	int                ans_len;
	struct xcamac_data     *xcamac = file->private_data;

	udev   = xcamac->dev;
	pipe   = usb_sndbulkpipe(udev, xcamac->ep_out);

	r = usb_bulk_msg(udev, pipe,(unsigned char *)buf, len, &ans_len, HZ*10);
	if( r < 0 ){
		printk("send_data ERR=%d\n", r);
		return -1;
	}
	return ans_len; 
}

/*--------------------------------------------------------------------------*/
static unsigned int
scsi_write(struct file *file, unsigned char *cdb, unsigned short cdb_size, void *buf, int len)
{
	int     r;
	int     rest_len, cur_len;

	DSWRITE(printk("*** in scsi_write\n"));
	if( (r=send_cdb(file, cdb, cdb_size)) < 0 ){
		return -EFAULT;
	}

	rest_len = len;
	while( rest_len > 0 ){
		cur_len = ( MAX_TRANS_BYTES < rest_len )
			? MAX_TRANS_BYTES : rest_len ;

		DSWRITE(printk("*** cur_len = %x\n", cur_len));
		if( (r=send_data(file, buf, cur_len )) < 0 ){
			return r;
		}
		buf      += cur_len;
		rest_len -= cur_len;
	}
	DSWRITE(printk("*** out scsi_write\n"));
	return len;
}

/*--------------------------------------------------------------------------*/
static int 
check_camac_qx(struct file *file)
{
	struct xcamac_data     *xcamac = file->private_data;
	DCHKCAM(printk("** in check_camac_qx\n"));

	scsi_request_sense(file);
	DCHKCAM(printk("** out check_camac_qx\n"));

	if( xcamac->sense_key  == 0x09 && 
	    xcamac->sense_code == 0x80 && 
	    xcamac->sense_qual == 0x06 )
		return 1;
	else if( xcamac->sense_key  == 0x09 && 
		 xcamac->sense_code == 0x80 && 
		 xcamac->sense_qual == 0x05 )
		return 0;
	else if( xcamac->sense_key  == 0x00 && 
		 xcamac->sense_code == 0x00 && 
		 xcamac->sense_qual == 0x00 )
		return 3;
	else
		return 0; 

}
/*--------------------------------------------------------------------------*/
static int
xcamac_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
#ifdef TIMEMEASURE
	struct timeval tv1, tv2;
	long sec, usec;
#endif
	struct xcamac_data     *xcamac;
  
	DXIOCTL(printk("** in ioctl\n"));

	xcamac = file->private_data;

	if( xcamac->dev == NULL ){
		printk("** pdata->dev = NULL\n");
		return -ENODEV;
	}

	if(xcamac->is_busy++){
		printk("** EBUSY\n");
		printk("%x\n", xcamac->is_busy++);
		return -EBUSY;
	}

	switch (cmd) {

	case INQUIRY:
	{
		int i;
		size_t len = 0x31; 
		unsigned char buf[0x31];
		unsigned short cdb_size = 0x06;
		unsigned char cdb_cam_inq[0x06] = {
			/* 0x00 */ INQUIRY_3929,/* INQUIRY COMMAND */
			/* 0x01 */ 0x00,        /* must 0 */
			/* 0x02 */ 0x00,        /* must 0 */
			/* 0x03 */ 0x00,        /* must 0 */
			/* 0x04 */ 0x31,        /* INQIRY returned data size */
			/* 0x05 */ 0x00         /* must 0 */
};      
		DXIOCTL(printk("** INQUIRY\n"));
    
		if( scsi_read(file, cdb_cam_inq, cdb_size, (unsigned int *)buf, len ) != len ){
			printk("** scsi_read fail\n");
			return -1;
		}

		for(i = 8; i < len; i++)
			printk("%c", buf[i]);
		printk("\n");

		break;
	}

	case BLOCKL:
	{
		size_t word_size = 4, len, func; 
		unsigned short cdb_size = 0x0a;
		unsigned int *word_buf, rlen;
		unsigned char cdb_cam_blk[0x0a] = {
			/* 0x00 */ BLOCK_3929,  /* CAMAC BLOCK COMMAND */
			/* 0x01 */ 0x00,        /* must 0 */
			/* 0x02 */ 0x20,        /* Mode Cont. */
			/* 0x03 */ 0x00,        /* NAF Hi */
			/* 0x04 */ 0x00,        /* NAF Lo */
			/* 0x05 */ 0x00,        /* Trans. Byte Count HI */
			/* 0x06 */ 0x00,        /* Trans. Byte Count MID */
			/* 0x07 */ 0x00,        /* Trans. Byte Count LO */
			/* 0x08 */ 0x00,
			/* 0x09 */ 0x00
		};
		
		struct camac_block *cam_blk;
		cam_blk = kmalloc(sizeof(struct camac_block), GFP_KERNEL);
		if(cam_blk == NULL ){
			xcamac->is_busy = 0;
			return -ENOMEM;
		}

		DXIOCTL(printk("** BLOCKL\n"));
		copy_from_user(cam_blk, (void *)arg, sizeof(struct camac_block));
		DXIOCTL(printk("cam_blk size = %x\n", sizeof(cam_blk)));

		len = cam_blk->ch * word_size;
		//printk("ch = %x\n", cam_blk->ch);
		//printk("len = %x\n",len);

		word_buf = kmalloc(len, GFP_KERNEL);
		if(word_buf == NULL ){
			xcamac->is_busy = 0;
			return -ENOMEM;
		}
		//if(len > 2048 ) return -1;

		cdb_cam_blk[0x02] |= cam_blk->mode;
		cdb_cam_blk[0x03] = (unsigned char)((cam_blk->naf & 0xff00) >>8);
		cdb_cam_blk[0x04] = (unsigned char)( cam_blk->naf & 0x00ff);
		cdb_cam_blk[0x05] = (unsigned char)((len & 0xff0000) >> 16);
		cdb_cam_blk[0x06] = (unsigned char)((len & 0x00ff00) >> 8);
		cdb_cam_blk[0x07] = (unsigned char)( len & 0x0000ff);
		func = cam_blk->naf & 0x1f;

		if(func < 8)
			rlen = scsi_read(file, cdb_cam_blk, cdb_size, word_buf, len );
		else
			rlen = scsi_write(file, cdb_cam_blk, cdb_size, word_buf, len );

		cam_blk->retlen = rlen;

		if( rlen != len ){
			DXIOCTL(printk("scsi_read fail\n"));
			check_camac_qx(file);
			copy_to_user((void*)arg ,cam_blk, sizeof(struct camac_block));
			xcamac->is_busy = 0;
			return -1;
		}

		xcamac->sense_key  = 0x00;
		xcamac->sense_code = 0x00;
		xcamac->sense_qual = 0x00; 
		copy_to_user((unsigned int *)cam_blk->data, word_buf, len);
		copy_to_user((void*)arg ,cam_blk, sizeof(struct camac_block)); 

		kfree(word_buf);
		kfree(cam_blk);

		break;
	}

	case BLOCKW:
	{
		size_t word_size = 2, len; 
		unsigned short cdb_size = 0x0a;
		unsigned short *word_buf;
		unsigned int func,rlen;
		unsigned char cdb_cam_blk[0x0a] = {
			/* 0x00 */ BLOCK_3929,  /* CAMAC BLOCK COMMAND */
			/* 0x01 */ 0x00,        /* must 0 */
			/* 0x02 */ 0x20,        /* Mode Cont. */
			/* 0x03 */ 0x00,        /* NAF Hi */
			/* 0x04 */ 0x00,        /* NAF Lo */
			/* 0x05 */ 0x00,        /* Trans. Byte Count HI */
			/* 0x06 */ 0x00,        /* Trans. Byte Count MID */
			/* 0x07 */ 0x00,        /* Trans. Byte Count LO */
			/* 0x08 */ 0x00,
			/* 0x09 */ 0x00
		};
#ifdef TIMEMEASURE
		struct timeval tv1, tv2;
		long sec, usec;
		float fsec;
#endif

		struct camac_block *cam_blk;
		cam_blk = kmalloc(sizeof(struct camac_block), GFP_KERNEL);
		if(cam_blk == NULL ){
			xcamac->is_busy = 0;
			return -ENOMEM;
		}
		DXIOCTL(printk("** BLOCKW\n"));
		copy_from_user(cam_blk, (void *) arg, sizeof(struct camac_block));

		len = cam_blk->ch * word_size;
      
		word_buf = kmalloc(len, GFP_KERNEL);
		if(word_buf == NULL ){
			xcamac->is_busy = 0;
			return -ENOMEM;
		}
		//printk("len = %x\n",len);
		//if(len > 1024 ) return -1;

		cdb_cam_blk[0x02] |= cam_blk->mode;
		cdb_cam_blk[0x03] = (unsigned char)((cam_blk->naf & 0xff00) >>8);
		cdb_cam_blk[0x04] = (unsigned char)( cam_blk->naf & 0x00ff);
		cdb_cam_blk[0x05] = (unsigned char)((len & 0xff0000) >> 16);
		cdb_cam_blk[0x06] = (unsigned char)((len & 0x00ff00) >> 8);
		cdb_cam_blk[0x07] = (unsigned char)( len & 0x0000ff);
		func = cam_blk->naf & 0x1f;

#ifdef TIMEMEASURE
		do_gettimeofday(&tv1);
#endif

		if(func < 8)
			rlen = scsi_read(file, cdb_cam_blk, cdb_size, word_buf, len );
		else
			rlen = scsi_write(file, cdb_cam_blk, cdb_size, word_buf, len );

		cam_blk->retlen = rlen;

#ifdef TIMEMEASURE
		do_gettimeofday(&tv2);
		sec = (tv2.tv_sec) - (tv1.tv_sec);
		usec = (tv2.tv_usec) - (tv1.tv_usec);

		fsec = (float)sec + (float)usec/(float)1000000.;
		//printk("time = %f sec\n", fsec);
		printk("sec  = %ld sec\n", sec);
		printk("usec = %ld sec\n", usec);
#endif

		if( rlen != len ){
			check_camac_qx(file);
			copy_to_user((void*)arg ,cam_blk, sizeof(struct camac_block)); 
			DXIOCTL(printk("scsi_read fail\n"));
			xcamac->is_busy = 0;
			return -1;
		}

		xcamac->sense_key  = 0x00;
		xcamac->sense_code = 0x00;
		xcamac->sense_qual = 0x00; 
		copy_to_user((unsigned short *)cam_blk->data, word_buf, len);
		copy_to_user((void*)arg, cam_blk, sizeof(struct camac_block));

		kfree(word_buf);
		kfree(cam_blk);

		break;
	}

	case CAMACW:
	{
		size_t len = 2;
		unsigned short cdb_size = 0x06;
		unsigned int func;
		unsigned short buf;
		struct camac_single cam_sing;

		unsigned char cdb_cam_single[0x06] = {
			/* 0x00 */ SINGLE_3929, /* CAMAC SINGLE COMMAND */
			/* 0x01 */ 0x00,        /* must 0 */
			/* 0x02 */ 0x00,        /* Mode Cont. 16bit: 0x02, 24bit: 0x00*/
			/* 0x03 */ 0x00,        /* NAF Hi */
			/* 0x04 */ 0x00,        /* NAF Lo */
			/* 0x05 */ 0x00         /* must 0 */
		};
		
#ifdef TIMEMEASURE
		struct timeval tv1, tv2;
		long sec, usec;
		float fsec;
#endif

		DXIOCTL(printk("** CAMACW\n"));

		copy_from_user(&cam_sing, (void *) arg, sizeof(cam_sing)); 
		cdb_cam_single[0x02] = (unsigned char) 0x02;
		cdb_cam_single[0x03] = (unsigned char)((cam_sing.naf & 0xff00) >>8);
		cdb_cam_single[0x04] = (unsigned char)( cam_sing.naf & 0x00ff);
		func = cam_sing.naf & 0x1f;

		cam_sing.qx = 0;

		if( func < 8){
			DXIOCTL(printk("** CAMACW read\n"));
			cam_sing.data = 0;

#ifdef TIMEMEASURE
			do_gettimeofday(&tv1);
#endif
			if( scsi_read(file, cdb_cam_single, cdb_size, &buf, len ) != len ){
				printk("scsi_read fail\n");
				cam_sing.qx = check_camac_qx(file);
				copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing)); 
				xcamac->is_busy = 0;
				return -1;
			}
#ifdef TIMEMEASURE
			do_gettimeofday(&tv2);
			//sec = (tv2.tv_sec) - (tv1.tv_sec);
			usec = (tv2.tv_usec) - (tv1.tv_usec);
			printk("*** scsi_read usec = %ld usec\n", usec);
#endif
			xcamac->sense_key  = 0x00;
			xcamac->sense_code = 0x00;
			xcamac->sense_qual = 0x00; 
			cam_sing.data = buf & 0xffff;
			cam_sing.qx = 3;
			copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing));
	
			break;
		}
		else if(func > 15 && func < 24){
			DXIOCTL(printk("** CAMACW write\n"));
			buf = (unsigned short)(cam_sing.data & 0xffff);
			if( scsi_write(file, cdb_cam_single, cdb_size, &buf, len ) != len ){
				printk("scsi_write fail\n");
				cam_sing.qx = check_camac_qx(file);
				copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing));
				xcamac->is_busy = 0;
				return -1;
			}
			xcamac->sense_key  = 0x00;
			xcamac->sense_code = 0x00;
			xcamac->sense_qual = 0x00; 
			cam_sing.qx = 3;
			copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing)); 
		}
		else {
			DXIOCTL(printk("** CAMACW ctl\n"));
			if( scsi_no_dat(file, cdb_cam_single, cdb_size) != 0 ){
				//printk("scsi_no_dat fail\n");
				cam_sing.qx = check_camac_qx(file);
				copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing));
				xcamac->is_busy = 0;
				return -1;
			}
			xcamac->sense_key  = 0x00;
			xcamac->sense_code = 0x00;
			xcamac->sense_qual = 0x00; 
			cam_sing.qx = 3;
			copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing)); 
		}
      
		break;
	}

	case CAMACL:
	{
		size_t len = 4;
		unsigned int buf, func;
		struct camac_single cam_sing;
		unsigned short cdb_size = 0x06;
		unsigned char cdb_cam_single[0x06] = {
			/* 0x00 */ SINGLE_3929, /* CAMAC SINGLE COMMAND */
			/* 0x01 */ 0x00,        /* must 0 */
			/* 0x02 */ 0x00,        /* Mode Cont. 16bit: 0x02, 24bit: 0x00*/
			/* 0x03 */ 0x00,        /* NAF Hi */
			/* 0x04 */ 0x00,        /* NAF Lo */
			/* 0x05 */ 0x00         /* must 0 */
		};
    
		DXIOCTL(printk("** CAMACL\n"));
		copy_from_user(&cam_sing, (void *) arg, sizeof(cam_sing)); 

		//printk("cam_info.naf = %x\n",cam_sing.naf );
		cdb_cam_single[0x02] = (unsigned char)0x00;
		cdb_cam_single[0x03] = (unsigned char)((cam_sing.naf & 0xff00) >>8);
		cdb_cam_single[0x04] = (unsigned char)( cam_sing.naf & 0x00ff);
		func = cam_sing.naf & 0x1f;
		cam_sing.qx = 0;

		if( func < 8){
			DXIOCTL(printk("** CAMACL read\n"));
			cam_sing.data = 0x0;

			if( scsi_read(file, cdb_cam_single, cdb_size, &buf, len ) != len ){
				printk("scsi_read fail\n");
				cam_sing.qx = check_camac_qx(file);
				copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing));
				xcamac->is_busy = 0;
				return -1;
			}

			xcamac->sense_key  = 0x00;
			xcamac->sense_code = 0x00;
			xcamac->sense_qual = 0x00; 
			cam_sing.data = buf & 0xffffff;
			DXIOCTL(printk("in ioctl : cam_sing.data = %x\n", cam_sing.data));

			cam_sing.qx = 3;
			copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing)); 
		}      
		else if(func > 15 && func < 24){
			DXIOCTL(printk("** CAMACL write\n"));
			buf = cam_sing.data & 0xffffff;
			if( scsi_write(file, cdb_cam_single, cdb_size,  &buf, len ) != len ){
				printk("scsi_write fail\n");
				cam_sing.qx = check_camac_qx(file);
				copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing));
				xcamac->is_busy = 0;
				return -1;
			}

			xcamac->sense_key  = 0x00;
			xcamac->sense_code = 0x00;
			xcamac->sense_qual = 0x00; 
			cam_sing.qx = 3;
			copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing)); 
		}
		else {
			DXIOCTL(printk("** CAMACL control\n"));
			if( scsi_no_dat(file, cdb_cam_single, cdb_size) != 0 ){
				//printk("scsi_no_dat fail\n");
				cam_sing.qx = check_camac_qx(file);
				copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing));
				xcamac->is_busy = 0;
				return -1;
			}

			xcamac->sense_key  = 0x00;
			xcamac->sense_code = 0x00;
			xcamac->sense_qual = 0x00; 
			cam_sing.qx = 3;
			copy_to_user((void*)arg ,&cam_sing, sizeof(cam_sing)); 
		}

		break;
	}
	
	default:
		return -ENOIOCTLCMD;
		break;
	}

	DXIOCTL(printk("** out ioctl\n"));
  
	xcamac->is_busy = 0;
	return 0;
}
/*--------------------------------------------------------------------------*/
static int
xcamac_open(struct inode *inode, struct file *file)
{

	struct xcamac_data *xcamac;

	//lock_kernel();

	if( MINOR(inode->i_rdev) != USB_XCAMAC_MINOR_BASE ){
		return -ENODEV;
	}

	xcamac = global_xcamac;
	if( xcamac == NULL ){
		printk("illegual xcamac\n");
		return -ENODEV;
	}


	DXOPEN(printk("open: %x present: %x\n", xcamac->is_open, !xcamac->present));
	if(xcamac->is_open || !xcamac->present) {
	  //unlock_kernel();
		printk("OPEN: xcamac is busy\n");
		return -EBUSY;
	}

	xcamac->is_open = 1;
	xcamac->is_busy = 0;

	file->private_data = xcamac;
	MOD_INC_USE_COUNT;

	//unlock_kernel();

	return 0;
}
/*--------------------------------------------------------------------------*/
static int
xcamac_close(struct inode *inode, struct file *file)
{
	struct xcamac_data *xcamac;
	xcamac = (struct xcamac_data *) file->private_data;

	xcamac->is_open = 0;


	//lock_kernel();
	//if( xcamac->dev == NULL ){
	//  kfree(xcamac);
	//  global_xcamac = NULL;
	//}
	//unlock_kernel();

	MOD_DEC_USE_COUNT;

	return 0;
}
/*--------------------------------------------------------------------------*/
static struct file_operations
usb_xcamac_fops = {
        owner:          THIS_MODULE,
	ioctl:          xcamac_ioctl,
        open:           xcamac_open,
        release:        xcamac_close,
};
/*--------------------------------------------------------------------------*/
static int                         probe_count = 0;
/*--------------------------------------------------------------------------*/
static int
send_scsi_id_by_dev( struct usb_device *udev, int id )
{
        unsigned int       pipe;
        __u8               requesttype = 0x40;
        __u8               request     = 0x05;
        __u16              value       = 0x0107;
        __u16              index       = 0x0000;
        __u16              size        = 0;
        int                timeout     = 3*HZ; /* 5sec */
        int                r;

        pipe   = usb_rcvctrlpipe(udev, 0);
        DSSNDID(printk("send_scsi_id id = %d\n", id));
        value = ( id << 8 ) | 7;
        r = usb_control_msg(udev, pipe, request, requesttype, value,
                                                index, NULL, size, timeout);
	DSSNDID(printk("send_scsi_id r = %d\n", r));
        if( r != 0 ){
                printk("send_scsi_id ERR=%d\n", r);
                return -1;
        }
        return 0;
}
/*--------------------------------------------------------------------------*/
static int xcamac_read_proc(char *page, char **start, off_t off,
                          int count, int *eof, void *data)
{
        char *out = page;
	int len;
	
	
        out += sprintf (out, "Device open    : %x\n", global_xcamac->is_open);
	out += sprintf (out, "SCSI ID        : %x\n", global_xcamac->scsi_id);
	out += sprintf (out, "Sense Key      : %x\n", global_xcamac->sense_key);
	out += sprintf (out, "Sense Code     : %x\n", global_xcamac->sense_code);
	out += sprintf (out, "Sense Qualifier: %x\n", global_xcamac->sense_qual);
	
        len = out - page;
        len -= off;
        if (len < count) {
                *eof = 1;
                if (len <= 0) return 0;
        } else
                len = count;

        *start = page + off;

        return len;
}
/*--------------------------------------------------------------------------*/
static void *probe_xcamac(struct usb_device *dev, unsigned int ifnum,
			  const struct usb_device_id *id)
{
        struct usb_interface_descriptor   *if_desc;
        int                                num_cf;
        int                                num_if;
        int                                num_ep;
        unsigned long                      vid;
        unsigned long                      pid;
        unsigned long                      vid_pid;
        int                                m_class;
        int                                s_class;
        int                                proto;
        int                                ep0;
        int                                ep1;
        char                              *v = "UNKNOWN";


        vid = dev->descriptor.idVendor;
        pid = dev->descriptor.idProduct;

        vid_pid = ( vid << 16 ) | pid;

	DPROBE(printk("in probe_xcamac\n"));
	DPROBE(printk("vid_pid: %lx\n", vid_pid));

        switch( vid_pid ){
        case VID_PID_LOGITEC: v="LOGITEC"; break;

        default:
                printk("ERR: Unknown Vender ID vid=0x%04lX(%lu) pid=0x%04lX(%lu)\n",
                                        vid, vid, pid, pid );
		return NULL;
        }

        //printk("usb-xvme: %s driver\n", v);
        //dbg_msg(_("compiled on [%s %s]\n", __DATE__, __TIME__)); 

        if( probe_count != 0 ){
                printk("too many device not support \n"); 
                return NULL;
        }
        ++probe_count;
        DPROBE(printk("probe counts: %x\n", probe_count));

        num_cf = dev->descriptor.bNumConfigurations;
        num_if = dev->config[0].bNumInterfaces;

	DPROBE(printk("num_cf = %d\n",num_cf));
	DPROBE(printk("num_if = %d\n",num_if));

        if( !( num_cf == 1 && num_if == 1 ) ){
                printk("num_conf=%d num_if=%d\n", num_cf, num_if);
		return NULL;
        }

        if_desc = &dev->actconfig->interface[ifnum].altsetting[0];
        m_class = if_desc->bInterfaceClass;
        s_class = if_desc->bInterfaceSubClass;
        proto   = if_desc->bInterfaceProtocol;
        num_ep  = if_desc->bNumEndpoints;


	DPROBE(printk("m_class = %x\n", m_class));
	DPROBE(printk("s_class = %x\n", s_class));
	DPROBE(printk("proto = %x\n", proto));
	DPROBE(printk("num_ep = %x\n", num_ep));
	/*
        if( !( m_class == USB_CLASS_VENDOR_SPEC && s_class == 1 ) ){
                err_msg(_("illegual class (%d:%d)\n", m_class, s_class ));
                return NULL;
        }
	*/
        if( !( proto == 1 || num_ep == 2 ) ){
                printk("illegual prot or num_ep\n");
                return NULL;
        }

        /*
         * global_pdata alloc and initialize
         */
	global_xcamac = kmalloc(sizeof(*global_xcamac), GFP_KERNEL);
        if(global_xcamac == NULL ) {
                printk("no memory!");
                return NULL;
        }

	global_xcamac->is_open = 0;
	global_xcamac->is_busy = 0;
        global_xcamac->present = 1;
	global_xcamac->sense_key = 0;
	global_xcamac->sense_code = 0;
	global_xcamac->sense_qual = 0;
        init_waitqueue_head(&global_xcamac->wait_queue);

        /*
         * setiup ep_inp and ep_out
         */
        ep0 = if_desc->endpoint[0].bEndpointAddress;
        ep1 = if_desc->endpoint[1].bEndpointAddress;

	DPROBE(printk("ep0 = %x\n",ep0));
	DPROBE(printk("ep1 = %x\n",ep1));

        if( (ep0 & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN ){
                if( (ep1 & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN ){
                        printk("illegual ep0");
                        kfree(global_xcamac);
                        global_xcamac = NULL;
                        return NULL;
                }
                global_xcamac->ep_inp = ep0;
                global_xcamac->ep_out = ep1;
        }else{
                if( (ep1 & USB_ENDPOINT_DIR_MASK) != USB_DIR_IN ){
                        printk("illegual ep1");
                        //kfree(global_pdata);
                        //global_pdata = NULL;
                        kfree(global_xcamac);
                        global_xcamac = NULL;
                        return NULL;
                }
                global_xcamac->ep_inp = ep1;
                global_xcamac->ep_out = ep0;
        }

        /*
         * set_configuration
         */
	/*
        if( usb_set_configuration(dev, dev->config[0].bConfigurationValue) ) {
                printk("usb_set_configuration error");
		kfree(global_xcamac);
		global_xcamac = NULL;
                return NULL;
        }
        DPROBE(printk("vid=0x%04lX(%lu) pid=0x%04lX(%lu)\n", vid, vid, pid, pid));
	*/

        global_xcamac->dev = dev;
	
        //if( send_scsi_id_by_dev(dev, SCSI_ID) < 0 ){
        if( send_scsi_id_by_dev(dev, scsi_id) < 0 ){
	  return NULL;
        }

	global_xcamac->scsi_id = scsi_id;

        //printk("xvme : Please set SCSI-ID=%d\n",scsi_id );
        //printk("xvme : Please exec 'mknod /dev/xcamac c %d %d'\n",
	//         USB_MAJOR, USB_XCAMAC_MINOR_BASE );

        return global_xcamac;
}
/*--------------------------------------------------------------------------*/
static void
disconnect_xcamac(struct usb_device *dev, void *ptr)
{
        struct xcamac_data *xcamac = (struct xcamac_data *)ptr;

	if(xcamac->is_open) {
	  xcamac->is_open = 0;
	  xcamac->dev = NULL;
	}
	xcamac->present = 0;
	//global_xcamac = NULL;
	kfree(global_xcamac);
	global_xcamac = NULL;

        if( probe_count != 1 ){
                printk("illegual probe_count=%d\n", probe_count );
        }

        --probe_count;
        info("XCAMAC disconnected");
}

/*--------------------------------------------------------------------------*/
static struct usb_driver
driver_table = {
        name:       "xcamac",
        probe:      probe_xcamac,
        disconnect: disconnect_xcamac,
        fops:       &usb_xcamac_fops,
        minor:      USB_XCAMAC_MINOR_BASE,
};
int __init
usb_xcamac_module_init(void)
{
	create_proc_read_entry ("driver/xcamac", 0, 0, xcamac_read_proc, NULL);
	return usb_register(&driver_table) < 0 ? -1 : 0;
}
/*--------------------------------------------------------------------------*/
void __exit
usb_xcamac_module_cleanup(void)
{
	remove_proc_entry ("driver/xcamac", NULL);
	usb_deregister(&driver_table);
}
/*--------------------------------------------------------------------------*/
MODULE_AUTHOR("Kazuo Nakayoshi ");
MODULE_DESCRIPTION("USB Device Driver for LUB-SC and K3929 ");
/*--------------------------------------------------------------------------*/
module_init (usb_xcamac_module_init);
module_exit (usb_xcamac_module_cleanup);
/*--------------------------------------------------------------------------*/

